> module Yahc.Interaction where
> import Yahc.Derivation (Der, step, conclusion, mkder, addStep, undoStep, result, steps, lhsDer)
> import Yahc.Expr (QExpr, unlift, equiv)
> import Yahc.Rules (parseRule, showRules, Axiom(prop), rules, axiom, name)
> import Yahc.Pretty (Pretty, render, pretty)
> import Yahc.Parser (parseExp)
> import Yahc.Command 
> import Yahc.Error
> import Yahc.Utils
> import Yahc.Checker

> import qualified Yahc.Unicode as U
> import Control.Monad
> import Control.Monad.State
> import Control.Arrow
> import Data.Maybe (maybe, isNothing)
> import Data.Char (isSpace, toUpper)
> import System.Exit (exitWith, ExitCode(..))
> import System.IO 
> -- import System.Console.Readline


> exitSuccess :: IO ()
> exitSuccess = exitWith ExitSuccess

> undoC = Command { cmdname = Undo
>                 , action = undoAction
>                 , longStr = "undo"
>                 , short = "u"
>                 , help = helpUndo
>                 }

> quitC = Command { cmdname = Quit
>                 , action = lift . const exitSuccess
>                 , longStr = "quit"
>                 , short = "q"
>                 , help = "Sale del programa"
>                 }

> helpC = Command { cmdname = Help
>                 , action = helpAction
>                 , longStr = "help"
>                 , short = "h"
>                 , help = helpHelp
>                 } 


> showC = Command { cmdname = Show
>                 , action = showAction
>                 , longStr = "show"
>                 , short = "sh"
>                 , help = helpShow
>                 } 

> nextC = Command { cmdname = Next
>                 , action = nextAction
>                 , longStr = "next"
>                 , short = "n"
>                 , help = helpNext
>                 }

> startC = Command { cmdname = Start
>                  , action = startAction
>                  , longStr = "start"
>                  , short = "s"
>                  , help = helpStart
>                  }

> addRuleC = Command { cmdname = AddRule
>                    , action = addRuleAction
>                    , longStr = "add"
>                    , short = "a"
>                    , help = helpAddRule
>                    }

> commands :: [Command]
> commands = [ quitC
>            , helpC
>            , showC
>            , addRuleC
>            , nextC
>            , startC
>            , undoC
>            ]
             

> putStrLift :: Action
> putStrLift = lift . putStrLn . U.toUTF8

> putStrPretty :: Pretty a => a -> ChkState ()
> putStrPretty = putStrLift . render .  pretty 

> putStrExpr :: QExpr -> ChkState ()
> putStrExpr = putStrPretty . unlift

> putStrErr :: Action
> putStrErr err = gets derState >>= 
>                 maybe (putStrLift err)
>                       (\der -> maybe (return ())
>                        (putStrLift . ("Error en la línea "++) . show) (lineNo der) >>
>                        (putStrLift err)
>                       )

> emptyDer :: ChkState ()
> emptyDer = gets derState >>= \state -> when (isNothing state) (putStrLift errEmptyDer)


> showDerStep :: DerState -> Axiom -> QExpr -> ChkState ()
> showDerStep der rule res = putStrPretty $ mkder (lhs der) [step rule res]
>     where lhs = conclusion . derivation

> showStep :: Axiom -> QExpr -> ChkState ()
> showStep rule res = putStrPretty $ step rule res


> endDer :: ChkState ()
> endDer = guardState (return ()) (\der -> 
>               putStrPretty (derivation der) >>
>               putStrLift "Ha concluido la derivación." >>
>               maybe (return ()) (putStrLift . ("Líneas leídas: " ++) . show) (lineNo der)
>               )

> showDer :: ChkState ()
> showDer = guardState emptyDer (putStrPretty . derivation)

> showThms :: ChkState ()
> showThms = gets theorems >>= mapM_ putStrLift . showRules 

> startAction :: Action
> startAction params = do exps <- liftM splitParams (return params)
>                         unless (length exps < 2) $ do
>                           [e1,e2] <- return exps
>                           parseExpL e1 (\from -> parseExpL e2 
>                                   (\goal -> setState (mkder from []) (return goal) >> 
>                                            putStrExpr from ))

> nextAction :: Action
> nextAction params = guardState emptyDer (\der -> do 
>                      parts <- liftM splitParams (return params)
>                      unless (length parts < 2) $ do
>                        [ruleName, expr] <- return parts
>                        parseRuleL ruleName (\rule -> 
>                          parseExpL expr (\res -> 
>                            liftedIf (liftM not (addStepSt rule res)) (return ()) 
>                                     (maybe (return ()) 
>                                      (\g -> if res == g then endDer else showDerStep der rule res)
>                                      (goal der))))
>                          )

> addRuleAction :: Action
> addRuleAction params = guardState emptyDer (\der -> 
>                       if (null . steps . derivation) der
>                       then putStrErr errNoStep
>                       else do (s,t) <- (return . (lhsDer &&& result . last . steps) . derivation) der
>                               parts <- return $ splitParams params
>                               unless (length parts < 2) $ do
>                                 [ruleName, ruleAbr] <- return parts
>                                 (return (axiom ruleName (s `equiv` t) [ruleAbr] ruleName) >>=
>                                  modify . addRule >>
>                                  putStrLift ("Se agregó la regla "++ ruleName)))

> showRule :: Action
> showRule name = parseRuleL name (putStrExpr . prop)

> showAction :: Action
> showAction param = case param of
>                   "rules" -> showThms
>                   "derivation" -> showDer
>                   "goal" -> guardState emptyDer ((maybe (return ()) putStrExpr) . goal)
>                   what -> if take 4 what == "rule" then showRule (drop 5 what)
>                          else putStrLift helpShow

> helpAction :: Action
> helpAction = maybe (putStrLift helpHelp) putStrLift . liftM help . readCmd commands
  
> undoAction :: Action
> undoAction = const $ guardState emptyDer (const undoStepState)

> parseExpL :: String -> (QExpr -> ChkState ()) -> ChkState ()
> parseExpL e next = liftM parseExp (return e) >>= either (putStrLift . show) next 

> parseRuleL :: String -> (Axiom -> ChkState ()) -> ChkState ()                    
> parseRuleL r next = gets theorems >>= 
>                     liftM (flip parseRule r) . return >>=
>                     maybe (putStrErr errNoSuchRule) next


> interaction :: ChkState ()
> interaction = lift (readline "> ") >>= 
>               maybe (return "") return >>= 
>               (\l -> if null l then return () else 
>                     return (parseCommand commands l) >>=
>                     maybe (putStrErr errMissingCmd) (\ m -> addLine l m >>= uncurry ($)) >>
>                     putStrLift "") >>
>               interaction
>     where addLine l m = return m -- lift (addHistory l) >> return m
>           readline s = putStr s >> hFlush stdout >> getLine >>= return . return


> readLineY :: Handle -> ChkState String
> readLineY handle = liftedIf (lift (hIsEOF handle))
>                    (lift (return ""))
>                    (gets derState >>= 
>                     maybe (return ()) (\d -> modify (\c -> c { derState = return (nextLine d)})) >> 
>                     lift (hGetLine handle) >>= 
>                     \line -> if commented line
>                             then readLineY handle
>                             else return line)
>     where commented :: String -> Bool
>           commented =  (=="--") . take 2

> compFile :: Handle -> ChkState ()
> compFile h = do lineRule <- readLineY h
>                 lineRes  <- readLineY h
>                 liftedIf (return (null lineRule)) 
>                          (endDer >> lift (hClose h) >> lift exitSuccess)
>                          (parseRuleL lineRule 
>                                      (\ax -> parseExpL lineRes (\res -> 
>                                              liftedIf (liftM not (addStepSt ax res))
>                                                       (putStrErr (errBadStep ++ name ax) >> 
>                                                        lift exitSuccess)
>                                                       (compFile h)))
>                          )
                       
> checkHandle :: String -> Handle -> ChkState ()
> checkHandle name handle = do 
>   atEOF <- lift $ hIsEOF handle
>   unless atEOF $ do
>     exprL <- readLineY handle
>     parseExpL exprL (\expr ->
>                      setStateF expr 1 name >>
>                      compFile handle)

> checkFile :: FilePath -> ChkState ()
> checkFile fname = do 
>   handle <- lift $ openFile fname ReadMode
>   checkHandle fname handle

> checkStdIn :: ChkState () 
> checkStdIn = checkHandle "stdin" stdin

> welcome, welcomeInt :: String
> welcome =  "yahc: Versión: 0.1"
> welcomeInt = "Para ver los posibles comandos y lo que hacen podes hacer \"help\""
